CON Touch
'' ILI9325 driver using the Touchblade161 design for faster ram to display transfers 
'' Average Joe and James Moxham 2012

  Margin = 4                                            ' some characters have xoffset of -1 or -2 eg v, and won't display on the extreme left edge
  _1ms      = 1_000_000 /     1_000                         ' Divisor for 1 ms

  ' touchscreen pins
  Touch_Clock = 0                        
  'Touch_ChipSelect = done with a group select
  Touch_DataIn = 1            
  Touch_DataOut = 2' touchscreen pins

    ' kye sd numbers
  _dopin = 24
  _clkpin = 25
  _dipin = 26
  _cspin = 27
  _cdpin = -1 ' -1 if unused.
  _wppin = -1 ' -1 if unused.
  _rtcres1 = -1 ' -1 always.
  _rtcres2 = -1 ' -1 always.  
  _rtcres3 = -1 ' -1 always.  
  _statuspin = -1 ' Status LED pin number.

VAR
   byte FileNumber                                      ' 0 to 31
   byte DisplayMode                                     ' type of display 0 = ILI9325, 1= SSD1289, 2 etc
   word ScreenWidth                                     ' either 240 in portrait or 320 in landscape
   word ScreenHeight                                    ' 320 in portrait or 240 in landscape
   long orientation                                     ' true is portrait
   long curx, cury     
   byte FontHeight              ' size of the current loaded font (pixels = fontsize *.75)
   word BackFontColor              ' the background color in RRRRRGGG GGGBBBBB format
   byte Latch                   ' current latch output
   long StringAddress                               ' start of string address
   'byte i2cbyte
   byte latchbyte
   long wcboot                  ' most recent boot warm or cold                                          ' 


OBJ
   spi:           "SPI_ASM"                               ' thanks to Beau - spi driver for touch screen    
   fat:           "SD-MMC_FATEngine.spin"                   ' Kye's latest sd driver and no RTC    
                                                                                              
DAT 
    RamLocation         long $0[31]                  ' 32 file ram locatinos
    RamSize             long $0[31]                  ' File sizes in words - NOT bytes (divide byte by 2)
    sdbuffer            byte $0[1024] ' 1024 byte buffer for sd card interface and also for sending one row 320 x3 = 960 bytes max picture size 
    buffer2             byte $0[512]  ' 512 general purpose hub buffer 
    rambuffer           word $0[256]  ' 512 byte (256 word) buffer easier for using with words

PUB BeginDesktop(DisplayType)                           ' store the type of display once by the desktop program
  DisplayMode := DisplayType                            ' global variable 0 = ILI9325, 1=SSD1289    
  Begin(true)                                           ' start up everything running the desktop
  result := wcboot

PUB BeginProgram                                        ' used by all programs except the desktop program
  Begin(false)                                          ' any program except the desktop
  result := wcboot

PRI Begin(Desktop) |n                                   ' desktop true if run from the desktop program, false for other programs
  start_ram_cog                                         ' * start cog first *start the external ram / display driver in a cog, needed for many routines so always first
  Latch_AllHighStart                                    ' all Latch pins high and enables pin 22
  Load161(0)                                            ' reset all the 161 counters to zero
  if desktop == false
     DisplayMode := GetDisplayMode                      ' reads "I" or "S" etc off the ram ** must have run desktop at least once first to store this *** 
     'DisplayMode := "I" ' temp for debugging, using the ILI9325   
  SelectMemGroup                                        ' select group2
  Latch_DisplayBoth                                     ' set Latch for both hand display so intialises both (one may not be present but does not matter)
  if displaymode == "I"
    Start_ILI9325                                         ' start the display 0 = ILI9325, 1 = SSD1289                 
  if displaymode == "S"
    Start_SSD1289
  wcboot := WarmColdBoot                                 ' get the last warm or cold boot value
  SetOrientation(true)                                  '  in portrait (true) or landscape mode (false), must be before clearscreen otherwise don't know screensize
  if wcboot == false and Desktop == true                ' no need to redisplay every time if worked the first time
    Text(string("SD"))                                    ' string to send, uses internal prop font so can see something if the SD fails to mount   
  fat.fatEngineStart( _dopin, _clkpin, _dipin, _cspin, _wppin, _cdpin, _rtcres1, _rtcres2, _rtcres3)
  fat.mountPartition(0)                                 ' mount the sd card
  if desktop == false                                   ' running a program, this is called without knowing what the ramlocation and ramsize values are
     ManualValues                                        ' coming from outside so reset ramsize and ramlocation values for 0 and 1
  SetCursor(0,0)                                        ' cursor for debugging to 0,0
  if desktop == false
    SelectSPIGroup                                        ' selects this group and starts the cog

'PUB BasicTesting
'   text(string("Basic testing")) 
'   sdbuffer[0] := 6
'   SpinHubToRam(@sdbuffer,5000,10) ' move data to ram
'   sdbuffer[0] := 7             ' change the value
'   SpinRamToHub(@sdbuffer,5000,10)  ' read back the original value
'   hex(sdbuffer[0],2)
'   'Drawline(100,100,110,100,$FFFF)       
'   Clearrectangle(120,120,129,129,$FFFF) 
'   Draw(100,100,109,109)                
'   repeat 100
'     pixel(%11111000_00000000)

     

' ************** string routines stored in external memory so can have arrays and large text files **********************

' to make things simpler, one array StringArray. any number of entries, each up to 128 characters
' pass the source and destination numbers
' TextArray(256) = 256 entries for text files etc
' first entry is destination, second entrie(s) are the source
' Commands are
' StringDim                     ' reserves memory 256*128, increments filenumber, adds filesize                                 
' StringLoadFile(Name,S)        ' read a .txt file off the SD card and stores in TextArray, new string entry is crlf
' StringSaveFile(Name,S,F)      ' save .txt file from TextArray to SD card adding in crlf, Start to Finish
' StringClear(Dest)             ' clear a string, pass string number
' StringAddChar(Dest,N)         ' adds a character to a string at end
' StringSubChar(Dest)           ' remove a character from the end
' StringCopy(Dest,Src)          ' copy string from one location to another
' StringStore(Dest,Src)         ' convert Prop string("mystring") to string stored in external memory
' StringLeft(Dest,Src,N)        ' Basic Left$
' StringMid(Dest,Src,L,N)       ' Basic Mid$
' StringRight                   ' Basic Right$
' StringInstr                   ' Basic Instr
' StringUcase(Dest,Src)         ' Basic Ucase$
' StringLcase                   ' Basic Lcase$
' StringLen                     ' Length
' StringStr                     ' string representation of a number
' StringHex(Dest,N)             ' number to hex string
' StringBin                     ' number to binary string
' StringVal                     ' value (hex is 0x, binary is 0b)
' StringJoin(Dest,Src1,Src2)    ' same as Basic + with strings
' StringInsert(Dest,Src1,Src2,n) ' inserts src2 into src1 starting at position n, (1 is first char) and moves to dest
' StringDebug(n)                ' print using the green propeller font string number n
' StringLen()                   ' same as strsize but works on external memory strings
' StringRambuffer               ' moves string to rambuffer array and returns result = location of this array
' StringCompare(Source1,Source2) ' returns true if strings equal in length and characters                 

' use these routines amongst others.

'integerToDecimal(number, length) '' 5 Stack Longs
'integerToHexadecimal(number, length) '' 5 Stack Longs
'integerToBinary(number, length) '' 5 Stack Longs
'decimalToInteger(characters) | sign '' 10 Stack Longs
'hexadecimalToInteger(characters) | sign '' 10 Stack Longs
'binaryToInteger(characters) | sign '' 10 Stack Longs

PUB StringTest ' test all the string routines
   'StringDim(10)
   
   'StringStore(5,string("Hello World"))
   'StringDebug(5)
   'StringClear(5)
   'StringAddChar(5,"1")
   'StringAddChar(5,"2")
   'StringAddChar(5,"3")
   'StringAddChar(5,"4")          
   'StringDebug(5)
   'StringSubChar(5)             ' remove one character from the right, useful for backspace etc
   'StringDebug(5)
   'StringCopy(7,5)              ' copy
   'StringDebug(7)
   'StringLeft(7,7,5)            ' source and destination the same or different
   'StringDebug(7)
   'StringMid(8,5,7,3)           ' dest, source, starting byte, number of bytes. first byte is 1
   'StringDebug(8)
   'StringRight(3,5,4)           ' dest, source, number of bytes
   'StringDebug(3)
   'Hex(StringInstr(5,2,"o"),2)  ' find this character starting at 2
   'StringBin(2,255)             ' convert a number to binary always 32 + 2 characters
   'StringDebug(2)
   'StringHex(4,100)                 ' always 8 + 2 characters
   'StringDebug(4)
   'StringDec(9,-20)                  ' always 11 characters
   'StringDebug(9)
   'StringStr(9,-7)                     ' Basic str$ removes leading zeroes, and first character is either - or a space
   'StringDebug(9)
   'StringStore(5,string("65"))                ' test string number to number
   'hex(StringVal(5),8)                             
   'StringStore(5,string("0x1234"))            ' test string hex to number
   'hex(StringVal(5),8)
   'StringStore(5,string("0b101"))         ' test string binary to number
   'hex(StringVal(5),8)
   'StringStore(1,string("Hello "))
   'StringStore(2,string("World"))
   'StringJoin(1,1,2)                            ' string1 = string1 +string2 (shows dest and a source can be the same - or could be different)
   'StringDebug(1)
   'StringUcase(1,1)                         ' Ucase test
   'StringDebug(1)
   'StringLcase(1,1)                         ' Lcase test
   'StringDebug(1)
   'StringStore(1,string(" 123"))             ' test the trim function
   'StringTrim(1,1)
   'StringDebug(1)
   'StringStore(1,string("TEST6.TXT"))          ' file name
   'StringStore(2,string("Line a of text"))      ' store some lines of text
   'StringStore(3,string("Line b of text"))
   'StringStore(4,string("Line c of text"))
   'StringSavefile(1,2,4)                       ' filename is 1, store lines 2 to 4
   'StringStore(1,string("TEST6.TXT"))             ' file name 
   'StringLoadFile(1,2)                        ' read back the file created above. 
   'Stringdebug(2)
   'Stringdebug(3)
   'Stringdebug(4)
   'repeat

PUB StringDim(n)                                ' reserve space for a string array, n= number of strings, 128 max bytes per string
   Ramsize[Filenumber] := n << 6                        ' * 64 as this is in words, not bytes
   StringAddress := NextLocation                          ' get the location of the string array  
   Filenumber += 1

PUB StringStore(StringN,Source)                               ' store a propeller string("test") to array number Dest 
   bytemove(@rambuffer, Source, (strsize(Source) + 1))     ' move to rambuffer array
   StringToRam(StringN)

PUB StringDebug(StringN)                                ' print using the green propeller font - useful if sd card not working so no fonts loaded
   RamToString(StringN)
   Text(@rambuffer)                                     ' print it out

PUB StringClear(StringN)
   bytefill(@rambuffer,0,128)                            ' only really need to do the first one but may help debugging to completely clear the string
   StringToRam(StringN)

PUB StringAddChar(StringN,c) | s                        ' add c to the string
   if c >31 and c <127                                  ' a character that can be displayed
     RamToString(StringN)                               ' get the string
     s := strsize(@rambuffer)
     byte[@rambuffer][s] := c       
     byte[@rambuffer][s + 1] := 0                       ' add in the new zero (can't assume array is cleared)
     StringToRam(StringN)                               ' store back to ram

PUB StringSubChar(StringN)                              ' remove one character from the right same as strings.left(strings.len - 1)   
   RamToString(StringN)
   if strsize(@rambuffer) <> 0                          ' if at least one character
     byte[@rambuffer][strsize(@rambuffer)-1] :=0
   StringToRam(StringN)

PUB StringCopy(Dest,Source)                             ' copy string from source to dest
   RamToString(Source)
   StringToRam(Dest)

PUB StringLeft(Dest,Source,n)                           ' Basic Left$
   RamToString(Source)
   byte[@rambuffer][n] := 0                             ' move the terminator
   StringToRam(Dest)                                    ' move back to ram

PUB StringMid(Dest,Source,start,n) | i                          ' Basic Mid$
   RamToString(Source)
   repeat i from 0 to n-1
     byte[@rambuffer][i] := byte[@rambuffer][i+start-1]       ' possibly could use bytemove here
   byte[@rambuffer][n] := 0                                'add in the new zero terminator  
   StringToRam(Dest)

PUB StringRight(Dest,Source,n) | i,m
   RamToString(Source)
   m := strsize(@rambuffer) - n 
   repeat i from 0 to n                                 ' Basic Right$. include the terminator
     byte[@rambuffer][i] := byte[@rambuffer][i+m]
   StringToRam(Dest)

PUB StringInstr(Source,Start,c) | i                              ' find c in the string starting at start
    RamToString(Source)
    result := 0                                              ' return 0 if not found
    repeat i from (start-1) to strsize(@rambuffer)
      if byte[@rambuffer][i] == c
        result := i+1                                   ' 1 is first character
        quit   
     
PUB StringBin(Dest,n)                                   ' convert n to a binary value at Dest. Prefix 0b
    repeat result from 33 to 2
      byte[@rambuffer][result] := ((n & 1) + "0")
      n >>= 1
    byte[@rambuffer][34] := 0                           ' add zero terminator
    byte[@rambuffer][0] := "0"                          ' prefix 0b same as some versions of C
    byte[@rambuffer][1] := "b"                          
    StringToRam(Dest)                                   ' store

PUB StringHex(Dest,n)                                   ' number to hex, always 8 characters
  repeat result from 9 to 2
    byte[@rambuffer][result] := lookupz((n & $F): "0".."9", "A".."F")
    n >>= 4
  byte[@rambuffer][10] := 0                              ' zero at end
  byte[@rambuffer][0] := "0"                          ' prefix 0x same as C
  byte[@rambuffer][1] := "x"   
  StringToRam(Dest)

PUB StringDec(Dest,n) | sign                                   ' number to decimal string, always 11 characters
  sign := "+"
  if(n < 0)
    sign := "-"
  if(n == negx)
    bytemove(@rambuffer, string("-2147483648"), 11)           ' max negative number special case
  else
    repeat result from 10 to 1
      byte[@rambuffer][result] := ((||(n // 10)) + "0")
      n /= 10
  byte[@rambuffer][0] := sign                           ' sign at the front
  byte[@rambuffer][11] := 0                             ' zero terminator
  StringToRam(Dest)

PUB StringStr(Dest,n) | found                                    ' same as Basic STR - similar to Dec above but sign is either space or negative, and leading zeros removed  
   StringDec(Dest,n)                                    ' number preserved in rambuffer
   found := 10                                          ' if answer was zero don't remove leading zero
   repeat result from 1 to 10
     if byte[@rambuffer][result] <> "0"
       found := result
       quit
   repeat result from 1 to 11
     byte[@rambuffer][result] := byte[@rambuffer][result +found -1]             ' shuffle over if needed    
   if byte[@rambuffer][0] == "+"
     byte[@rambuffer][0] := " "                         ' Basic syntax, - or blank
   StringToRam(Dest)                                    ' resend to external ram

PUB StringVal(Source) | sign,i,k                        ' returns value of the string, hex 0x, binary 0b, decimal + or - or space or no leading character just a number
  RamToString(Source)
  if byte[@rambuffer][1] == "x"                         ' this is a hex number
    repeat k from 2 to strsize(@rambuffer) -1           ' leading two characters always 0x
      result := ((result <- 4) + (byte[@rambuffer][k] & $F))  
    return

  if byte[@rambuffer][1] == "b"                         ' this is a binary number
     repeat k from 2 to strsize(@rambuffer) -1           ' leading two characters always 0b 
       result := ((result << 1) + (byte[@rambuffer][k] & 1))  
     return

  ' else if got here this must be a decimal number. It could start with a number(no leading sign), or a space, or a + or a -
  sign := byte[@rambuffer][0]
  i := 0                                                ' start location
  if sign == "+" or sign == "-" or sign == " "           ' this is a decimal number ** must start with a space or a +  or a -**
    i := 1                                               ' leading character so start here
  repeat k from i to (strsize(@rambuffer)-1)              ' if ever use right to left, don't use step -1, spin puts it in if start is less than finish and if there is a step -1 it is one less loop
    result := ((result *10) + byte[@rambuffer][k] & $F)   'very clever, the & $F works because zero is hex0x30, the pre multiply shifts previous numbers to the left so can do left to right. Easier to understand with Kye's binary converter with the left shift
  if sign == "-"
    result *= -1                                       ' if negative then negate

PUB StringJoin(Dest,Source1,Source2)                    ' Dest$= Source1$+Source2$
   RamToString(Source2)
   bytemove(@buffer2,@rambuffer,128)                    ' store string2 to buffer2
   RamToString(Source1)                                 ' get first string
   bytemove(@rambuffer+strsize(@rambuffer),@buffer2,strsize(@buffer2)+1)          ' thanks to Kye for this one
   StringToRam(Dest)                                    ' store back to ram

PUB StringUcase(Dest,Source)
   RamToString(Source)
   repeat result from 0 to strsize(@rambuffer)
     if ((byte[@rambuffer][result] => "a") and (byte[@rambuffer][result] =< "z"))
       byte[@rambuffer][result] -= 32
   StringToRam(Dest)

PUB StringLcase(Dest,Source)
   RamToString(Source)
   repeat result from 0 to strsize(@rambuffer)
     if ((byte[@rambuffer][result] => "A") and (byte[@rambuffer][result] =< "Z"))
       byte[@rambuffer][result] += 32
   StringToRam(Dest)

PUB StringTrim(Dest,Source)
   RamToString(Source)
   if byte[@rambuffer][0] == "+" or  byte[@rambuffer][0] == " " or byte[@rambuffer][0] == "-"         
     repeat result from 0 to strsize(@rambuffer)
       byte[@rambuffer][result] := byte[@rambuffer][result+1]      ' delete the first +,- or space
   StringToRam(Dest)

PUB StringSaveFile(Filen,Start,Finish) | i,j ' filen might be string array 0, start could be 1, finish could be 4        
   RamToString(Filen)                                 ' fetch the filename
   fat.newfile(@rambuffer)        ' create the file, if not working use \fat.  for debugging, see PUB openfileread for an example
   result := \fat.openfile(@rambuffer,"W")                        ' open for writing return error if there is one
   repeat i from Start to Finish
     RamToString(i)
     j := strsize(@rambuffer)
     byte[@rambuffer][j] := 13                          ' carriage return
     byte[@rambuffer][j+1] :=10                         ' line feed
     byte[@rambuffer][j+2] := 0                          ' end of new line
      fat.writedata(@rambuffer,j+2)                      ' write this line
   FileClose                                            ' close the file

PUB StringLoadFile(Filen,Stringnumber)| s,r,c,t,z ' read filename, new line every crlf. Must have enough space reserved in the string array
    RamToString(Filen)                                                                                              ' copy a file from the SD card to the ram disk
    OpenFileRead(@rambuffer)
    s := 0                                              ' sd counter
    r := 0                                              ' rambuffer counter
    t := 0                                              ' total number of bytes
    z := fat.filesize                                   ' size of the file
    repeat (z >> 9) +1                                  ' include any remainder as +1, do blocks of 512 bytes each as natural size of SD blocks
      fat.readdata(@sdbuffer,512)                       ' read from SD card
      repeat s from 0 to 511
        c := byte[@sdbuffer][s]                         ' get the next byte
        if t =< z                                       ' might read in one more 512 byte block to ensure gets any remainder
          if c == 13 or r > 126                         ' end of a line so new string, or string too long
            byte[@rambuffer][r] := 0                    ' end of string terminator
            StringToRam(Stringnumber)
            stringnumber +=1                            ' new string number
            r := 0                                      ' start at beginning of rambuffer
          if c >31 and c < 127
            byte[@rambuffer][r] := c                    ' store if a valid character
            r +=1
        t +=1                                           ' number of bytes read
    FileClose                                       ' close the file

PUB StringLen(Source)                                   ' string length
   RamToString(Source)
   result := strsize(@rambuffer)

PUB StringPrint(Font,Source) | i,s                      ' print this string using using font in memory and curx and cury
  RamToString(Source)
  s := strsize(@rambuffer)
  repeat i from 0 to (s-1)
    curx := TextChar(Font,byte[@rambuffer][i],curx,cury)

PUB StringRamBuffer(Source)                             ' string to rambuffer array
   RamToString(Source)
   result := @rambuffer                                 ' return pointer to the rambuffer array

PUB StringCompare(Source1,Source2) | i
   RamToString(Source2)
   bytemove(@buffer2,@rambuffer,128)                    ' store string2 to buffer2
   RamToString(Source1)
   result := true
   repeat i from 0 to strsize(@rambuffer)
     if byte[@buffer2][i] <> byte[@rambuffer][i]
       result := false
       return    
      


PRI StringToRam(StringN)
   HubToRam(StringN << 6 + StringAddress,64)                        ' 64 words =128 bytes, so work in groups of 64 words

PRI RamToString(StringN)    
   RamToHub(StringN << 6 + StringAddress,64)                          ' moves to rambuffer array    

' ***************** end string routines *************************

PUB DisplayLeft                                         ' send all data to the left display
    Latch_Display2

PUB DisplayRight
    Latch_Display1       

                                  
PUB RamErase(n)
    Filenumber := n                                     ' erase lookup table for the ram files, effectively erases the files    
                                                        ' if 0 then erase all, if 1 then leave the desktop intact (assuming this was loaded first)

PUB ClearRam
    Filenumber := 2                                     ' clear ram but leave file 0 = warm or cold boot and 1 = background picture

PUB GetCurX
    result := curx                                       ' get the text cursor position

PUB GetCurY
    result := cury

PUB SetCurx(n)                                            ' set the text cursor position
    Curx := n

PUB SetCury(n)
    Cury := n

PUB GetBackFontColor
   result := BackFontColor

PUB GetRamLocation(n)
   result := Ramlocation[n]

PUB GetFilenumber
   result := filenumber

PUB SetFilenumber(n)                                     ' reset the file number
    Filenumber := n  

PUB Chain(stringptr)
    result := \fat.bootpartition(stringptr) ' reboot and run this program            

PUB SDtoRam(stringptr) | RamAddress,remainder,sizebytes,sizewords                              ' copy a file from the SD card to the ram disk
    OpenFileRead(stringptr)
    sizebytes := fat.filesize                           ' size in bytes
    sizewords := sizebytes >>1                          ' size in words
    ramaddress := NextLocation                          ' next free place in ram
    RamSize[FileNumber] := sizewords                    ' ram disk works in words, not bytes
    repeat sizebytes >> 9                               ' do blocks of 512 bytes each as natural size of SD blocks
      fat.readdata(@sdbuffer,512)                       ' read from SD card
      SpinHubToRam(@sdbuffer,RamAddress,512)            ' move to ram
      RamAddress += 256                                 ' increment ram address by 256 (half 512 as using words)
    remainder :=  sizebytes & %00000000_00000000_00000001_11111111 ' remainder (if not a multiple of 512) 
    if remainder <> 0
      fat.readdata(@sdbuffer,remainder)
      SpinHubToRam(@sdbuffer,RamAddress,remainder)             ' send out the remainder bytes
    FileClose                                       ' close the file
    result := FileNumber                                ' return this file
    FileNumber +=1                                      ' point to next location

PUB FetchRamBuffer
    result := @rambuffer    

PUB SDBMPtoRam(stringptr) | width,height,w,i,ramaddress       ' .bmp from sd to ram and convert to 2 byte ili format and reverse order along the way
' http://en.wikipedia.org/wiki/BMP_file_format scroll down about 1/3 of the way for the header formats
' get the width, height
' store these as 2 longs at the beginning of the ram file
' the bitmap format starts at the last row and works up so need to reverse the order
    ramaddress := NextLocation                          ' get the next ram location
    OpenFileRead(stringptr)
    fat.readdata(@sdbuffer,$36)                         ' get the header 0x36 hex bytes
    width := sdbuffer[$12] + sdbuffer[$13] << 8         ' only read in two bytes as never will be >64k wide or high
    height := sdbuffer[$16] + sdbuffer[$17] << 8
    w := width * 3                                      ' this number of bytes per row, and round up to nearest 4 bytes et 327 goes to 328 The size of each row is rounded up to a multiple of 4 bytes (a 32-bit DWORD) by padding.
    w +=3                                               ' add 3 and
    w &= %11111111_11111111_11111111_11111100           ' round down
    SpinHubToRam(@sdbuffer,ramaddress,$36)         ' store header to ram
    i := $36 + ramaddress + ((height - 1) * width) ' start + header and on the last row and work back
    repeat height
      fat.readdata(@sdbuffer,w)                         ' read in the first row
      CogCmd("F",@sdbuffer,@rambuffer,width)
      HubToRam(i,width)                                 ' send to ram
      i -= width                                        ' subtract the width, move up the picture
    FileClose
    Ramsize[Filenumber] := $36 + (width * height)      ' size of bitmap file in words
    result := FileNumber                                ' returns the current file number
    FileNumber += 1                                     ' add 1 to filenumber

PUB DrawBMPRam(f,x,y) | rampointer,width,height          ' pass the filenumber and the x and y location, searches for this file and if found, displays it
   rampointer := RamLocation[f]                         ' find the start of the file
   SpinRamToHub(@sdbuffer,rampointer,$36)                      ' read in the header
   width := sdbuffer[$12] + sdbuffer[$13] << 8         ' only read in two bytes as never will be >64k wide or high
   height := sdbuffer[$16] + sdbuffer[$17] << 8
   Draw(x,y,x+width-1,y+height-1)                    ' set up the area to draw
   RamToDisplay(RamPointer + $36,width*height)       ' skip the header and display
   result := (width <<16) | (height & $0000FFFF)     ' return the width and height combined into a long

' ***************** end ram disk routines

PUB MergeBackgroundIconRam(stringptr,x,y,desktop,mask,icon) | i,j,k,n                                 ' merge icon and background - if mask =0 then background, else icon (not classic alpha compositing but close enough)   
     ' i is background counter, j is icon counter, k = mask counter
    ' assumes desktop is in ramfile 1, mask in ramfile 2, icon in ramfile 3
    FileNumber := icon 
    SDBMPtoRam(stringptr)                               ' load icon into ram  
    Filenumber := icon                                     ' reset filenumber
    i := RamLocation[desktop] + $36 + (y*(screenwidth+1)+x)   ' location of picture pixel in ram
    j := RamLocation[icon] + $36                           ' location of icon in pixel in ram
    k := RamLocation[mask] + $36                           ' location of mask pixel in ram
    repeat 60
      SpinRamToHub(@sdbuffer,i,59)                             ' get 59 pixels from background
      SpinRamToHub(@rambuffer,j,59)                            ' get 59 icon pixels
      SpinRamToHub(@buffer2,k,59)                              ' get words for the mask
      'repeat n from 0 to 58
      '  if (word[@buffer2][n] & %00000000_00011111) > %10000 ' RGB are the same, so use BBBBB, mask 5 low bits, and if > %10000
      '     word[@sdbuffer][n] :=word[@rambuffer][n]     ' if >128 then replace with icon pixel
      CogCmd("X",@sdbuffer,@rambuffer,@buffer2)              ' pasm version of the above 3 lines    
      SpinHubToRam(@sdbuffer,i,59)                             ' send back modified line to the picture buffer    
      i += screenwidth+1                              ' next line in picture
      j +=59                                            ' icon next line
      k +=59                                            ' next line in mask (bytes) - buffer2        

PUB OpenFileRead(stringptr) ' open a file for reading and display a message if it fails
   result := \fat.openfile(stringptr,"R")
   if fat.partitionerror <> 0                           ' failed to find the file
     Text(stringptr)                         ' print the filename
     Text(result)                            ' print the error message
     repeat                                             ' stop the program

PUB FileClose
    fat.closefile                                       ' close the file

PUB TouchXPercent
    result := TouchValue & 255                          ' decode xval 0-100%
    if DisplayMode == "S"
      result := 100-result                              ' SSD1289 x runs the other way


PUB TouchYPercent
    result := TouchValue >> 8                           ' decode yval 0-100%  

PUB TouchSwishX | xval,oldxval,difference,average ' returns result = +ve or -ve depending on swish direction     ' did finger move left or right?
      SelectSPIGroup ' turn on the spi touchscreen
      oldxval := 255' for startup
      average := 0
      repeat until (average > 30) or (average < -30)
        ' possibly add a decay function here for average
        pause1ms(50)
        xval := TouchXPercent                        ' decode xval 0-100%
        if xval == 255                                  ' not touched
          oldxval := 255                                ' reset oldxval as well
        else
          if oldxval <>255                                ' discard the first reading as the screen is touched
            difference := xval - oldxval
            'printdecimal(xval)
            'printdecimal(difference)
            average := average + difference             ' 
            'printdecimal(average)
            'crlf
          oldxval := xval                                 ' store for next time
      result := average                                 ' return negative or positive value

PUB TouchSwishY | yval,oldyval,difference,average ' returns result = +ve or -ve depending on swish direction     ' did finger move left or right?
      SelectSPIGroup ' turn on the spi touchscreen
      oldyval := 255' for startup
      average := 0
      repeat until (average > 30) or (average < -30)
        ' possibly add a decay function here for average
        pause1ms(50)
        yval := TouchYPercent                        ' decode xval 0-100%
        if yval == 255                                  ' not touched
          oldyval := 255                                ' reset oldyval as well
        else
          if oldyval <>255                                ' discard the first reading as the screen is touched
            difference := yval - oldyval
            average := average + difference             ' 
          oldyval := yval                                 ' store for next time
      result := average                                 ' return negative or positive value             

PUB TouchStart
      SPI.start(3,0)                                        ' delay,state, start clock high

PUB TextChar(FileN,ascii,x,y) | jump,size,width,height,ramaddress,xoffset,yoffset,xadvance   ' read out a character from a .ifn file stored in ram
  ' pass Fonttable = global
  ' moves to next line if off the end
    ramaddress := RamLocation[FileN]
    jump := ReadRamLong(ramaddress,(ascii << 2) + 256)   ' jump location = ascii *4 plus fonttable + 256
    SpinRamToHub(@buffer2,ramaddress+jump>>1,20)                ' read in the width height as a block = quicker than readramlong
    size := long[@buffer2][0]                           ' size in pixels of this character
    xadvance := long[@buffer2][5]                       ' amount to move to next character   
    'if size > 0 and ascii <> 32                         ' no need to print anything if size is zero or read more font data
    width := long[@buffer2][1]                        ' width
    height := long[@buffer2][2]                       ' height
    xoffset := long[@buffer2][3]                      ' xoffset to move
    yoffset := long[@buffer2][4]                      ' yoffset to move
    if (x+width -1 + xoffset) > screenwidth
        crlf                                            ' do a new line if it won't fit
        x := curx
        y := cury
    if ascii < 32
      xadvance := 0                                     ' add nothing if a non displaying character
    if ascii == 13                                  ' carriage return
        cr
        x := curx
    if ascii ==10                                   ' line feed (new line)
        lf
        y := cury
    if ascii > 32 ' space not displayed
      Draw(x+xoffset,y+yoffset,x+width-1+xoffset,y+height-1+yoffset)      ' draw on screen
      RamToDisplay(ramaddress+((jump+32)>>1),size)       ' move bytes from ram out to the display
    return x + xadvance    

PUB crlf
    cr
    lf

PUB cr                                                  ' carriage return, to left margin
    curx := Margin                                      ' some characters are -1 xoffset so won't display

PUB lf                                                  ' line feed, move up one
    cury += FontHeight
    if cury > screenheight
      ClearRectangle(0,0,screenwidth,screenheight,BackFontColor) 
      cury := 0                                         ' if off the bottom of the screen then clear screen start at the top (VT00 does scrolling)
    ' don't clear background here, do in calling routine
    'ClearFontBackground(0,cury,screenwidth,cury + FontHeight) ' clear background line to the background color ready to draw          

PUB TextT(FileN,stringptr)                 'print at curx,cury using file number in ramdisk
'' Print a zero-terminated string
  repeat strsize(stringptr)
    curx := TextChar(FileN,byte[stringptr++],curx,cury)

    


PUB SetCursor(x,y)               
    curx := x
    cury := y


PUB ReadRamLong(WordStart,Bytevalue)                   ' easier calculating
    SpinRamToHub(@result,Wordstart+Bytevalue>>1,2)
    return result

PUB LoadFont(stringptr) | filesize,i                ' read a .ifn file premade in angelcode bmfont and then the vb.net program
    SDtoRam(stringptr)                                  ' font to ram disk, adds one to filenumber
    SpinRamToHub(@rambuffer,RamLocation[Filenumber-1],128) ' read first 128 words =256 bytes back
    FontHeight := byte[@rambuffer][251]                 ' font height
    repeat i from 0 to 2
      sdbuffer[i] := byte[@rambuffer][35+i]
    CogCmd("E",@sdbuffer,@buffer2,1)                         ' convert to 2 byte .ili format
    BackFontColor := ( byte[@buffer2][1] << 8) | byte[@buffer2][0] ' correct order
    result := FileNumber-1                              ' return pointer to the font file in ram


PUB HubToRam(RamAddress,Number)                         ' send pixels from rambuffer hub to ram
   SpinHubToRam(@rambuffer,RamAddress,Number)    ' spin version

PUB RamToHub(RamAddress,Number)                         ' send pixels from ram to hub at rambuffer
   SpinRamToHub(@rambuffer,Ramaddress, number)

PUB RamToDisplay(RamAddress,Number)                     ' send pixels from ram to display
   SpinRamToDisplay(Ramaddress, Number)

 '
'PUB Change161(RamAddress)                                ' pass ramaddress. Returns in group 1
'    docmd("Z",0,RamAddress,0)       

PUB Clearscreen(color)                                  ' uses 2 byte ILI color (see below for rgb conversion)
    wordfill(@rambuffer,color,256)
    Draw(0,0,screenwidth,screenheight)                           ' set up the area of the screen to draw in   
    repeat 300
      SpinHubToDisplay(@rambuffer,256)

PUB RGBtoWord(R,G,B)                                     ' convert RGB to 2 byte ILI color 
    sdbuffer[0] := R
    sdbuffer[1] := G
    sdbuffer[2] := B                        
    CogCmd("E",@sdbuffer,@buffer2,1)
    result := ( byte[@buffer2][1] << 8) | byte[@buffer2][0] ' correct order 
    
PUB ClearRectangle(x1,y1,x2,y2,ColorWord) | i,size,remainder ' clears an area of the screen to the current font background color
    size := (x2-x1+1) * (y2 - y1 +1)                    ' number of pixels
    remainder := size & 255                             ' if not a whole number of 256 pixels
    repeat i from 0 to 255                              ' move the background font colour to the buffer
      rambuffer[i] := ColorWord
    Draw(x1,y1,x2,y2)                                   ' set up the area of the screen to draw in
    repeat size >> 8                                    ' 256 pixel blocks
      SpinHubToDisplay(@rambuffer,256)                    
    if remainder <> 0  
      SpinHubToDisplay(@rambuffer,remainder)                     ' do the remainder

PUB Drawline(x1,y1,x2,y2,ColorWord)                     ' draws a vertical or horizontal line
   ClearRectangle(x1,y1,x2,y2,ColorWord)                ' same as drawing a rectangle 
      
PUB SetOrientation(e) ' change orientation true = portrait. Changes global variable 'orientation' and also screen width and height
   orientation := e
   ChangeOrientation(orientation)
   if orientation
     screenwidth :=239
     screenheight :=319
   else
     screenwidth :=319
     screenheight :=239

PUB ChangeOrientation(n) ' pass true = portrait or false = landscape, changes global variable orientation in this object
' command $0001
'8.2.4. Driver Output Control (R01h)
'R/W RS D15 D14 D13 D12 D11 D10 D9 D8 D7 D6 D5 D4 D3 D2 D1 D0
'W 1 0 0 0 0 0 SM 0 SS 0 0 0 0 0 0 0 0
'SS: Select the shift direction of outputs from the source driver.
'When SS = 0, the shift direction of outputs is from S1 to S720
'When SS = 1, the shift direction of outputs is from S720 to S1.
'In addition to the shift direction, the settings for both SS and BGR bits are required to change the
'assignment of R, G, B dots to the source driver pins.
'To assign R, G, B dots to the source driver pins from S1 to S720, set SS = 0.
'To assign R, G, B dots to the source driver pins from S720 to S1, set SS = 1.

' command $0003
 'R/W RS D15 D14 D13 D12 D11 D10 D9  D8 D7  D6   D5   D4   D3 D2 D1 D0
   'W   1  TRI DFM 0   BGR 0   0   HWM 0  ORG 0    I/D1 I/D0 AM 0  0  0
   ' When AM = 0, the address is updated in horizontal writing direction.
   'When AM = 1, the address is updated in vertical writing direction.
   'I/D[1:0] Control the address counter (AC) to automatically increase or decrease by 1 when update one pixel
   ' display data.
    SelectMemGroup                                      
    orientation := n
    if displaymode == "I"                                         ' ili9325
      if orientation
        ' for origin top left and portrait mode
        Displaycmd($0001,%00000001_00000000)                           ' $0001,$0100 set SS and SM bit 0001 0100   portrait
        Displaycmd($0003,%00010000_00110000)                           ' $0003,$1030 set GRAM write direction and BGR=1. $0003 $1030  
      else
        ' for origin top right and landscape mode 
        Displaycmd($0001,%00000000_00000000)                           ' set SS and SM bit 0001 0000   landscape   
        Displaycmd($0003,%00010000_00111000)              ' landscape  $1028 = original but 1038 is correct - not mirror image


    if displaymode == "S"              ' SSD1289
      if orientation
            ' for origin top left and portrait mode
            Displaycmd($0001,$2B3F)                           ' $0001,$0100 set SS and SM bit 0001 0100   portrait
            Displaycmd($0011,$6070) '6070                          ' $0003,$1030 set GRAM write direction and BGR=1. $0003 $1030  
      else
            ' for origin top right and landscape mode 
            Displaycmd($0001,$4B3F)                           ' set SS and SM bit 0001 0000   landscape   
            Displaycmd($0011,$6838)                     ' landscape  $1028 = original but 1038 is correct - not mirror image


PUB SelectMemGroup                                      ' select group 1 for memory transfer
   spi.stop                                             ' and stop any other cogs here too if needed
   LatchGroup2
   DIRA |= %00000000_01011111_11111111_11111111         ' enable these pins for output
   OUTA |= %00000000_00011111_00000000_00000000         ' set /rd /wr /disp wr and 161 clock high

   
PUB SelectSPIGroup
   LatchGroup3                                            ' select group 2 for spi transfers
   DIRA &= %11111111_11111111_11111111_11000000   ' so no clashes with the cog using P0-P2 set these as inputs and P3-P5 
   TouchStart                                     ' start the touchscreen cog
   
   ' may need to set some pins here as inputs or outputs with a dira

PUB Text(stringptr)                          'print at curx,cury
  repeat strsize(stringptr)
    Propfont_out(byte[stringptr++])
  PropFontcrlf

PUB PropFontcrlf
    curx := 0
    cury += 32                                          ' new line at end of string
    if cury >319                                        ' bottom of screen so new screen
      curx:=0
      cury:=0

PUB Propfont_out(ascii) | address,pixels
    Draw(curx,cury,curx+15,cury+31)                     ' location to start drawing 
    address := $8000 + (ascii >> 1) << 7                ' get rom address
    repeat 32                                           ' 32 rows per character, split in two parts
      pixels := long[address]                           ' get rom font data
      pixels := pixels >> (ascii & 1)                   ' shift for odd characters
      repeat 16 ' 16 columns
        if pixels & 1
          Pixel(%00000111_11100000)                     ' foreground color RRRRRGGG_GGGBBBBB
        else
          Pixel(%00000000_00000000)                     ' background color 
        pixels := pixels >> 2                           ' alternate pixels interleaved so shift 2
      address += 4 
    curx +=16
    if curx >239                                        ' new line
      Propfontcrlf

PUB Start_ILI9325
    LatchGroup2                                           ' talk to this display
    LCD_Reset_High
    pause1ms(5)
    LCD_Reset_Low
    pause1ms(5)
    LCD_Reset_High                                      
    LCD_CS_High
    LCD_RD_High
    LCD_WR_High
    pause1ms(20)
    LCD_CS_Low
    ILI9325initiate

PUB Start_SSD1289                                       ' based on C driver
    LatchGroup2                                           ' talk to this display
    LCD_Reset_High  
    pause1ms(5) 
    LCD_Reset_Low      
    pause1ms(10) 
    LCD_Reset_High                                      
    LCD_CS_High
    LCD_RD_High
    LCD_WR_High
    pause1ms(20)
    SSD1289initiate

PUB Draw(x1, y1, x2, y2)                                ' sets the pixel to x1,y1 and then fills the next (x2-x1)*(y2-y1) pixels
    SelectMemGroup                                      ' select memory group
    ifnot orientation                                   ' landscape mode so swap x and y
      result :=x1                                       ' swap x1 and y1
      x1 := y1
      y1 := result
      result := x2                                      ' swap x2 and y2
      x2 :=y2
      y2 := result

    if displaymode == "I"                            ' ILI9325 display
      Displaycmd($0050,x1)
      Displaycmd($0052,y1)
      Displaycmd($0051,x2)
      Displaycmd($0053,y2)
      Displaycmd($0020,x1)
      Displaycmd($0021,y1)
      Lcd_Write_Com($0022)
      
    if displaymode == "S"                           ' SSD1289 display
      Displaycmd($0044,(x2<<8)+x1)           
      Displaycmd($0045,y1)
      Displaycmd($0046,y2)
      Displaycmd($004e,x1)
      Displaycmd($004f,y1)
      Lcd_Write_Com($0022)

    LCD_RS_High                                       ' after sending out commands, set RS high via latch

PUB Pixel(pixelcolor)               ' send out a pixel
     Lcd_Write_Data(pixelcolor)

PUB pause1ms(period) | clkcycles

'' Pause execution for period (in units of 1 ms).

  clkcycles := ((clkfreq / _1ms * period) - 4296) #> 381     ' Calculate 1 ms time unit
  waitcnt(clkcycles + cnt)                                   ' Wait for designated time

PUB hex(value, digits)   ' uses internal prop font and slow pixel drawing (for very basic debugging without an SD font)
  SelectMemGroup                               ' reselect mem group for display output for debugging  
'' Print a hexadecimal number
  propfont_out("O")
  propfont_out("x")
  value <<= (8 - digits) << 2
  repeat digits
    propfont_out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))
  PropFontcrlf

PUB SetWarmBoot
    long[@rambuffer][5] := $5761726D                    ' spell Warm in hex, start at 5 as zero gets corrupted
    byte[@rambuffer][10] := DisplayMode                 ' store the current display mode
    HubToRam(0,20)                                      ' send words to external ram (ramdiskstart is zero)
    ClearScreen(0)                                      ' so user sees something happen as there is a bit of a delay rebooting
    reboot

PUB WarmColdBoot ' store warm or cold boot  - reserve first 20 bytes in sram for interfacing with programs
    ' Ramdisk file 0 is the boot string Warm or Cold
    ' ramdisk file 1 is the background
    Ramtohub(0,20)         ' read words from the ram 
    if long[@rambuffer][5] == $5761726D
      result := true
    else
      result := false
    long[@rambuffer][5] := $FFFFFFFF                    ' erase previous warm/cold boot
    byte[@rambuffer][10] := DisplayMode                 ' store display mode
    HubToRam(0,20)                            ' send words to external ram (ramdiskstart is zero)
    RamSize[0] := 20                          ' reserve 20 bytes
    RamLocation[0] := 0
    Filenumber := 1

PRI GetDisplayMode
    Ramtohub(0,20)         ' read words from the ram
    result := byte[@rambuffer][10]               ' get the display mode

PUB ManualValues
     RamLocation[0] := 0
     RamSize[0] := $14                                   ' put in values manually - 20 bytes for warm boot
     RamLocation[1] := $14
     RamSize[1] := $12C36                                ' 76800 for the desktop picture
     FileNumber := 2                                     ' running a program not a desktop

PRI LatchGroup1
   latchbyte |= %00001111         ' set all  group pins high
   latchbyte &= %11111110         ' set this one low
   Set373Latch

PRI LatchGroup2
   latchbyte |= %00001111         ' set all high
   latchbyte &= %11111101         ' set this one low
   Set373Latch 

PRI LatchGroup3
   Latchbyte |= %00001111         ' set all high 
   Latchbyte &= %11111011         ' set this one low
   Set373Latch   

PRI LatchGroup4
   Latchbyte |= %00001111         ' set all high
   Latchbyte &= %11110111         ' set this one low
   Set373Latch   

PRI Latch_ResetLow
   Latchbyte &= %11101111            ' set this pin low
   Set373Latch   

PRI Latch_ResetHigh
   Latchbyte |= %00010000            ' set this pin high
   Set373Latch   

PRI Latch_Display1               ' display on the right
   Latchbyte |= %01100000
   Latchbyte &= %11011111            ' set this pin low
   Set373Latch  

PRI Latch_Display2              ' display on the left
   Latchbyte |= %01100000
   Latchbyte &= %10111111            
   Set373Latch  

PRI Latch_DisplayBoth           ' both displays  for debugging
   Latchbyte &= %10011111            
   Set373Latch           
            
PRI Latch_RS_Low
   Latchbyte &= %01111111            ' set this pin low
   Set373Latch 

PRI Latch_RS_High
   Latchbyte |= %10000000            ' set this pin high
   Set373Latch 

PRI Latch_AllHighStart               ' all pins high
   DIRA |= %00000000_01000000_00000000_00000000      ' pin 22 always an output
   OUTA &= %11111111_10111111_11111111_11111111      ' pin 22 is low and nothing else changes it except 373 output
   Latchbyte := %11111111
   Set373Latch 

PRI Load161(address) | Oldlatch
   Oldlatch := LatchByte
   LatchByte |= %01100000                               ' disable both displays while change 161 otherwise prints garbage as the /wr toggles
   LatchGroup1                                          ' select 161 group
   DIRA |= %00000000_00011111_11111111_11111111         ' 19 address lines and 2 control lines
   OUTA &= %11111111_11100000_00000000_00000000         ' set P0-P18 and also the 2 control lines low
   OUTA |= address                                      ' merge with the address
   OUTA |= %00000000_00010000_00000000_00000000         ' clock high p20
   OUTA |= %00000000_00001000_00000000_00000000         ' load high p19
   OUTA &= %11111111_11101111_11111111_11111111         ' clock low
   OUTA &= %11111111_11110111_11111111_11111111         ' load low
   OUTA |= %00000000_00010000_00000000_00000000         ' clock high p20
   OUTA |= %00000000_00001000_00000000_00000000         ' load high p19
   DIRA &= %11111111_11100000_00000000_00000000         ' float all the pins used pins for the group handover
   LatchByte := OldLatch                                ' restore latchbyte, the next group change will update it

PRI SpinHubToRam(hubaddress, ramaddress, number) | i ' P0-P15 are data lines, P16 is /RD, P17 is /WR
   Load161(ramaddress)                                  ' load the 161 counters
   LatchGroup2                                            ' group 2 is block data transfers
   CogCmd("S",hubaddress,ramaddress,number)             ' pasm alternative to code below
   DIRA &= %11111111_11100000_00000000_00000000         ' float all pins used

PRI SpinRamToHub(hubaddress,ramaddress,number) |i       ' P0-P15 are data lines, P16 is /RD, P17 is /WR 
   Load161(ramaddress)                                  ' load the 161 counters
   LatchGroup2                                            ' group 2 is block data transfers
   CogCmd("T",hubaddress,ramaddress,number)             ' pasm alternative to code below    
   DIRA &= %11111111_11100000_00000000_00000000         ' float all pins used 

PRI SpinRamToDisplay(ramaddress,number) | i         ' ramaddress, number of words (pixels), display = 1 or 2
   Load161(ramaddress)                                  ' load the 161 counters
   LatchGroup2                                            ' group 1 is block data transfers
   CogCmd("U",0,0,number)                               ' pasm alternative to code below, much faster
   DIRA &= %11111111_11100000_00000000_00000000         ' float all pins used

PRI SpinHubToDisplay(hubaddress,number)|i
   'LatchGroup2                                            ' not needed as always do a Draw first
   CogCmd("V",hubaddress,0,number)                      ' pasm version of code below
   DIRA &= %11111111_11100000_00000000_00000000         ' float all pins used

 ' ***********************************************************


PRI NextLocation                                         ' get the next free location in ram based on last files size
    if FileNumber == 0
       RamLocation[0] := 0                              ' work out start of this file based on last file size
    else   
       RamLocation[FileNumber] := RamLocation[FileNumber - 1] + RamSize[FileNumber - 1]  ' get location in ram
    result := RamLocation[FileNumber]


PRI TouchX
      SPI.SHIFTOUT(Touch_DataIn, Touch_Clock, 5, 8 , %1101_0000)  ' reads x from  500 to 3700  (off < 500 )
      result := SPI.SHIFTIN(Touch_DataOut,Touch_Clock,2, 12)

PRI TouchY      
      SPI.SHIFTOUT(Touch_DataIn, Touch_Clock, 5, 8 , %1001_0000)  ' reads y from 400 to 3800 (off > 3800 ) 
      result := SPI.SHIFTIN(Touch_DataOut,Touch_Clock,2, 12)  

PRI TouchValue | xval,yval,x,y ' returns % values 0-100% (or 255 for no touch)
      xval := TouchX
      yval := TouchY
      if (xval => 500) and (yval =< 3800)                       ' both have to be valid for a touch
        pause1ms(2)                               ' tiny delay then read again as the first reads may not have the correct value
        xval := TouchX
        yval := TouchY
        if (xval => 500) and (yval =< 3800)                 ' and if valid again, now more likely this value is correct
            x := xval - 440                                    ' scale 0-3360
            x /=  32                                     ' scale 0-100
            x <#= 100                                    ' max 100
            x #>=0                                       ' min 0
            x := 100 - x                                 ' so left edge on left
            y := yval -420                               ' scale 0-3180
            y /= 32                                      ' scale 0-100
            y <#= 100                                    ' max 100
            y #>=0                                       ' min 0
            y := 100-y                                   ' so 0 is top left     
            result := (y <<8) | x                       ' return 00000000_00000000_YYYYYYYY_XXXXXXXX   
      if (xval < 500) or (xval >3800)
        result := $0000FFFF                             ' invalid number(s)

PRI ILI9325initiate    
          ' ************* Start Initial Sequence **********
    Displaycmd($00E5,$78F0)                             ' set SRAM internal timing
    Displaycmd($0001,$0100)                             ' set SS and SM bit 0001 0100   portrait
    Displaycmd($0002,$0700)                             ' set 1 line inversion 
    Displaycmd($0003,$1030)                             ' set GRAM write direction and BGR=1. $0003 $1030
    Displaycmd($0004,$0000)                             ' Resize register 
    Displaycmd($0008,$0207)                             ' set the back porch and front porch 
    Displaycmd($0009,$0000)                             ' set non-display area refresh cycle ISC[3:0]  
    Displaycmd($000A,$0000)                             ' FMARK function  
    Displaycmd($000C,$0000)                             ' RGB interface setting   
    Displaycmd($000D,$0000)                             ' Frame marker Position   
    Displaycmd($000F,$0000)                             ' RGB interface polarity  
'        *************Power On sequence ****************//
    Displaycmd($0010,$0000)                             ' SAP, BT[3:0], AP, DSTB, SLP, STB 
    Displaycmd($0011,$0007)                             ' DC1[2:0], DC0[2:0], VC[2:0] 
    Displaycmd($0012,$0000)                             ' VREG1OUT voltage   
    Displaycmd($0013,$0000)                             ' VDV[4:0] for VCOM amplitude  
    Displaycmd($0007,$0001)
    pause1ms(50)                                        ' Dis-charge capacitor power voltage
    Displaycmd($0010,$1090)                             ' 1490//SAP, BT[3:0], AP, DSTB, SLP, STB   
    Displaycmd($0011,$0227)                             ' DC1[2:0], DC0[2:0], VC[2:0] 
    pause1ms(50)                                        ' delay
    Displaycmd($0012,$001F)                             ' 001C// Internal reference voltage= Vci;
    pause1ms(50)                                        ' delay   
    Displaycmd($0013,$1500)                             ' $1000//1400   Set VDV[4:0] for VCOM amplitude  1A00  
    Displaycmd($0029,$0027)                             ' $0012 //001a  Set VCM[5:0] for VCOMH  //$0025  0034  
    Displaycmd($002B,$000D)                             ' Set Frame Rate   000C       
    pause1ms(50)                                        ' delay   
    Displaycmd($0020,$0000)                             ' GRAM horizontal Address   
    Displaycmd($0021,$0000)                             ' GRAM Vertical Address    
'         ----------- Adjust the Gamma Curve ----------//
    Displaycmd($0030,$0000)
    Displaycmd($0031,$0707)
    Displaycmd($0032,$0307)
    Displaycmd($0035,$0200)
    Displaycmd($0036,$0008)
    Displaycmd($0037,$0004)
    Displaycmd($0038,$0000)
    Displaycmd($0039,$0707)
    Displaycmd($003C,$0002)
    Displaycmd($003D,$1D04)
'        ------------------ Set GRAM area ---------------//
    Displaycmd($0050,$0000)                             ' Horizontal GRAM Start Address
    Displaycmd($0051,$00EF)                             ' Horizontal GRAM End Address       
    Displaycmd($0052,$0000)                             ' Vertical GRAM Start Address    
    Displaycmd($0053,$013F)                             ' Vertical GRAM Start Address       
    Displaycmd($0060,$A700)                             ' Gate Scan Line     
    Displaycmd($0061,$0001)                             ' NDL,VLE, REV    
    Displaycmd($006A,$0000)                             ' set scrolling line      
'        -------------- Partial Display Control ---------/
    Displaycmd($0080,$0000)
    Displaycmd($0081,$0000)
    Displaycmd($0082,$0000)
    Displaycmd($0083,$0000)
    Displaycmd($0084,$0000)
    Displaycmd($0085,$0000)
'        //-------------- Panel Control -------------------//
    Displaycmd($0090,$0010)
    Displaycmd($0092,$0600)
    Displaycmd($0007,$0133)                             ' 262K color and display ON
    ChangeOrientation(true)                             ' default to portrait


PRI SSD1289initiate
{       ILIcmddelay($0000,$0001)      'Turn Oscillator on                                POR-$0000
        ILIcmddelay($0003,$A8A4)      'Power control (1)                                 POR-$6664
        ILIcmddelay($000C,$0000)      'Power control (2)                                 POR- ?
        ILIcmddelay($000D,$080C)      'Power control (3)                                 POR-$0009
        ILIcmddelay($000E,$2B00)      'Power control (4)                                 POR-$3200
        ILIcmddelay($001E,$00B0)      'Power control (5)                                 POR-$0029
        ILIcmddelay($0001,$4B3F)      'Driver output control,    *landscape?? $6B3F*     POR-$433F
        ILIcmddelay($0002,$0600)      'LCD drive AC control                              POR-$0400
        ILIcmddelay($0010,$0000)      'Sleep Mode                sleep mode off          POR-$0001
        ILIcmddelay($0011,$6070)      'Entry Mode,               *landscape? $4030*      POR-$6830  
        ILIcmddelay($0005,$0000)      'Compare Register (1)                              POR-$0000
        ILIcmddelay($0006,$0000)      'Compare Register (2)                              POR-$0000
        ILIcmddelay($0016,$EF1C)      'Horizontal Porch                                  POR-$EFC1
        ILIcmddelay($0017,$0003)      'Vertical Porch                                    POR-$0003
        ILIcmddelay($0007,$0033)      'Display Control                                   POR-$0000
        ILIcmddelay($000B,$D308)      'Frame cycle Control                               POR-$D308
        ILIcmddelay($000F,$0000)      'Gate Scan start position                          POR-$0000
        ILIcmddelay($0041,$0000)      'Vertical Scroll Control (1)                       POR-$0000
        ILIcmddelay($0042,$0000)      'Vertical Scroll Control (2)                       POR-$0000
        ILIcmddelay($0048,$0000)      'First Window Start                                POR-$0000
        ILIcmddelay($0049,$013F)      'First Window End                                  POR-$013F
        ILIcmddelay($004A,$0000)      'Second Window Start                               POR-$0000
        ILIcmddelay($004B,$0000)      'Second Window End                                 POR-$013F
        ILIcmddelay($0044,$EF00)      'Horizontal Ram Address Postion                    POR-$EF00
        ILIcmddelay($0045,$0000)      'Vertical Ram Address Start                        POR-$0000
        ILIcmddelay($0046,$013F)      'Vertical Ram Address End                          POR-$013F
        ILIcmddelay($0023,$0000)      'RAM write data mask (1)                           POR-$0000
        ILIcmddelay($0024,$0000)      'RAM write data mask (2)                           POR-$0000
        ILIcmddelay($0025,$8000)      'not in datasheet?                                     
        ILIcmddelay($004f,0)          'RAM Y address counter                             POR-$0000
        ILIcmddelay($004e,0)          'RAM X address counter                             POR-$0000
        Lcd_Write_Com($0022)
}        

' based on C code

    Displaycmd($0000,$0001)
    Displaycmd($0003,$A8A4)
    Displaycmd($000C,$0000)
    Displaycmd($000D,$080C)
    Displaycmd($000E,$2B00)
    Displaycmd($001E,$00B0)
    Displaycmd($0001,$2B3F)
    Displaycmd($0002,$0600)
    Displaycmd($0010,$0000)
    Displaycmd($0011,$6070)
    Displaycmd($0005,$0000)
    Displaycmd($0006,$0000)
    Displaycmd($0016,$EF1C)
    Displaycmd($0017,$0003)
    Displaycmd($0007,$0233)        
    Displaycmd($000B,$0000)
    Displaycmd($000F,$0000)
    Displaycmd($0041,$0000)
    Displaycmd($0042,$0000)
    Displaycmd($0048,$0000)
    Displaycmd($0049,$013F)
    Displaycmd($004A,$0000)
    Displaycmd($004B,$0000)
    Displaycmd($0044,$EF00)
    Displaycmd($0045,$0000)
    Displaycmd($0046,$013F)
    Displaycmd($0030,$0707)
    Displaycmd($0031,$0204)
    Displaycmd($0032,$0204)
    Displaycmd($0033,$0502)
    Displaycmd($0034,$0507)
    Displaycmd($0035,$0204)
    Displaycmd($0036,$0204)
    Displaycmd($0037,$0502)
    Displaycmd($003A,$0302)
    Displaycmd($003B,$0302)
    Displaycmd($0023,$0000)
    Displaycmd($0024,$0000)
    Displaycmd($0025,$8000)
    Displaycmd($004f,$0000)
    Displaycmd($004e,$0000)
    Lcd_Write_Com($0022)
    ChangeOrientation(true)                             ' default to portrait     

PRI Displaycmd(c,d) ' instruction in one method
    Lcd_Write_Com(c) ' send out a word
    Lcd_Write_Data(d)

PRI Lcd_Write_Com(LCDlong)
    LCD_RS_low                               ' can do rs first then cs - better for latch board
    LCD_CS_Low                               ' not needed for the ILI9325 but the SSD1289 needs this
    LCD_Writ_Bus(LCDlong)
    LCD_CS_High                              ' not needed for the ILI9325 but the SSD1289 needs this
 
PRI Lcd_Write_Data(LCDlong)
    LCD_RS_High                              ' can do rs first then cs
    LCD_CS_Low                               ' not needed for the ILI9325 but the SSD1289 needs this  
    LCD_Writ_Bus(LCDlong)
    LCD_CS_High                              ' not needed for the ILI9325 but the SSD1289 needs this  

PRI LCD_Writ_Bus(LCDLong)
    'LCDLong &= %00000000_00000000_11111111_11111111    ' mask to a word. Not needed if care taken always to send a word value
    OUTA &= %11111111_11111111_00000000_00000000        ' set P0-P15 to zero ready to OR
    OUTA |= LCDLong                                     ' merge with the word to output
    LCD_WR_Low                                          ' send out the data
    LCD_WR_High

PRI LCD_WR_Low
   OUTA &= %11111111_11111011_11111111_11111111         ' p18 low

PRI LCD_WR_High
   OUTA |= %00000000_00000100_00000000_00000000         ' p18 high

PRI LCD_RD_High                                         ' not used, set high in hardware

PRI LCD_CS_Low
   OUTA &= %11111111_11110111_11111111_11111111         ' p19 low    

PRI LCD_CS_High                                         
   OUTA |= %00000000_00001000_00000000_00000000         ' p19 high  

PRI LCD_RESET_Low                                       ' reset low
   Latch_ResetLow

PRI LCD_RESET_High
   Latch_ResetHigh

PRI LCD_RS_Low                                          ' for historical reasons this subroutine calls another
   Latch_RS_Low

PRI LCD_RS_High
   Latch_RS_High      


PRI Set373Latch                                     ' set latch to latchbyte value
'   CogCmd("N",LatchByte,0,0)
' spin version for debugging as the cog one is not working ? returning P22 and P8 in the wrong state
    DIRA |= %00000000_01000000_00000001_11111111        ' P0-7, P8 and P22 as outputs
    OUTA &= %11111111_11111111_11111110_11111111        ' P8 low
    OUTA |= %00000000_01000000_00000000_00000000        ' P22 high
    OUTA &= %11111111_11111111_11111111_00000000        ' P0-P7 low
    OUTA |= latchbyte                                   ' output the byte
    OUTA |= %00000000_00000000_00000001_00000000        ' P8 high
    OUTA &= %11111111_11111111_11111110_11111111        ' P8 low
    OUTA &= %11111111_10111111_11111111_11111111        ' P22 low     
   
   
' ************ cog drivers ***************

PRI start_ram_cog : err_
' Initialise the Drac Ram driver. No actual changes to ram as the read/write routines handle this
  command := "I"
  cog := 1 + cognew(@tbp2_start, @command)
  if cog == 0
    err_ := $FF                 ' error = no cog
  else
    repeat while command        ' driver cog sets =0 when done
    err_ := errx                ' driver cog sets =0 if no error, else xx = error code

PRI stop_ram_cog
   if cog
      cogstop(cog~ - 1)      

PRI CogCmd(command_, hub_address, ram_address, block_length) : err_| a,b
' Do the command: A-Z (I is reserved for Initialise)
  'dira := 0                     ' tristate all pins with the spin dira
  a := dira                     ' store the state of these and restore at the end 
  b := outa                     
  DIRA &= %11111111_11100000_00000000_00000000 ' tristate all common pins that cog can change so no conflicts
  ' p22 is complicated - need to change it within a cog to set 161 fast within cog, BUT if tristate it here
  ' then it will be pulled high and that will disable the latch and upset all the latch outputs
  ' need to think about this
  ' it should work if DIRA is set outside and is always an output, and then in spin and in pasm
  ' it can set the logic level of the pin.
  hubaddrs := hub_address       ' hub address start
  ramaddrs := ram_address       ' ram address start
  blocklen := block_length      ' block length
  command  := command_          ' must be last !!
' Wait for command to complete and get status
  repeat while command          ' driver cog sets =0 when done
  err_ := errx                  ' driver cog sets =0 if no error, else xx = error code
  dira := a                     ' restore dira and outa
  outa := b

CON
'' Modified code from Cluso's triblade
'' commands to move blocks of data to the ILI9325 touchscreen display
' DoCmd(command_, hub_address, ram_address, block_length)
' I - initialise     
' S - Move data from hub to ram
' T - Move data from ram to hub
' U - Move data from ram to display
' V - Hub to display
' E - convert from .raw RGB to two byte ILI format RRRRRGGG_GGG_BBBBB
' F - convert from .bmp BGR format to two byte ILI format
' X - merge icon and background based on a mask
' Z - Set 161 pins. Returns in group 1
' N - Set 373 latch

VAR

' communication params(5) between cog driver code - only "command" and "errx" are modified by the driver
   long  command, hubaddrs, ramaddrs, blocklen, errx, cog ' rendezvous between spin and assembly (can be used cog to cog)
'        command  = A to Z etc =0 when operation completed by cog
'        hubaddrs = hub address for data buffer
'        ramaddrs = ram address for data
'        blocklen = ram buffer length for data transfer
'        errx     = returns =0 (false=good), else <>0 (true & error code)
'        cog      = cog no of driver (set by spin start routine)
   

DAT
'' +-----------------------------------------------------------------------------------------------+
'' | Touchblade 161 Ram Driver (with grateful acknowlegements to Cluso and Average Joe)            |
'' +-----------------------------------------------------------------------------------------------+
                        org     0
tbp2_start    ' setup the pointers to the hub command interface (saves execution time later
                                      '  +-- These instructions are overwritten as variables after start
comptr                  mov     comptr, par     ' -|  hub pointer to command                
hubptr                  mov     hubptr, par     '  |  hub pointer to hub address            
ramptr                  add     hubptr, #4      '  |  hub pointer to ram address            
lenptr                  mov     ramptr, par     '  |  hub pointer to length                 
errptr                  add     ramptr, #8      '  |  hub pointer to error status           
cmd                     mov     lenptr, par     '  |  command  I/R/W/G/P/Q                  
hubaddr                 add     lenptr, #12     '  |  hub address                           
ramaddr                 mov     errptr, par     '  |  ram address                           
len                     add     errptr, #16     '  |  length                                
err                     nop                     ' -+  error status returned (=0=false=good) 


' Initialise hardware tristates everything and read/write set the pins
init                    mov     err, #0                  ' reset err=false=good
                        'mov     dira,zero                ' tristate the pins with the cog dira
                        and     dira,maskP0P20low       ' tristates all the common pins, leaves P22 as is though

done                    wrlong  err, errptr             ' status  =0=false=good, else error x
                        wrlong  zero, comptr            ' command =0 (done)
' wait for a command (pause short time to reduce power)
pause
'                        mov     ctr, delay      wz      ' if =0 no pause
'              if_nz     add     ctr, cnt
'              if_nz     waitcnt ctr, #0                 ' wait for a short time (reduces power)
                        rdlong  cmd, comptr     wz      ' command ?
              if_z      jmp     #pause                  ' not yet
' decode command
                        cmp     cmd, #"S"       wz      ' hub to ram
              if_z      jmp     #pasmhubtoram           
                        cmp     cmd, #"T"       wz      ' ram to hub
              if_z      jmp     #pasmramtohub
                        cmp     cmd, #"U"       wz      ' ram to display
              if_z      jmp     #pasmramtodisplay
                        cmp     cmd, #"V"       wz      ' hub to display
              if_z      jmp     #pasmhubtodisplay           
                        cmp     cmd, #"E"       wz      ' convert 3 byte .raw format to 2 byte .ili format - hub to hub
              if_z      jmp     #rawtoiliformat
                        cmp     cmd, #"F"       wz      ' convert 3 byte .bmp format BGR to 2 byte ili format (same as E but order reversed)
              if_z      jmp     #bmptoiliformat              
                        cmp     cmd, #"X"       wz      ' merge icon and background based on a mask
              if_z      jmp     #mergeicons
'                        cmp     cmd, #"Z"       wz     ' set the 161 counters
'              if_z      jmp     #set161
'                        cmp     cmd, #"N"       wz      ' set the 373 latch
'              if_z      jmp     #latch373           
                        cmp     cmd, #"I"       wz      ' init
              if_z      jmp     #init     
                        mov     err, cmd                ' error = cmd (unknown command)
                        jmp     #done
                        
' ----------------- common routines -------------------------------------

get_values              rdlong  hubaddr, hubptr         ' get hub address
                        rdlong  ramaddr, ramptr         ' get ram address
                        rdlong  len, lenptr             ' get length
                        mov     err, #5                 ' err=5
                        or      dira,maskP16P20         ' set control pins P16-P20 as outputs
                        or      outa,maskP16P20         ' set control pins high
get_values_ret          ret



stop                   jmp     #stop                  ' for debugging

delaynop                nop
                        nop
                        nop
                        nop
delaynop_ret            ret


' ------------------ single letter commands  -------------------------------------

' command S
pasmhubtoram            call    #get_values             ' get hubaddr,ramaddr,len and set control pins
                        or      dira,maskP0P15          ' %00000000_00000000_11111111_11111111         ' data bus outputs
hubtoram_loop           and     outa,maskP16P31         '%11111111_11111111_00000000_00000000       ' clear for output                   
                        rdword  data_16,hubaddr         ' get the word from hub
                        and     data_16,maskP0P15       ' mask to a word only
                        or      outa,data_16            ' send out the byte to P0-P15
                        andn    outa,maskP17            ' set mem write low
                        add     hubaddr,#2              ' increment by 2 bytes = 1 word. Put this here for small delay while writes
                        or      outa,maskP17            ' mem write high
                        andn    outa,maskP20            ' clock 161 low
                        or      outa,maskP20            ' clock 161 high
                        djnz    len,#hubtoram_loop      ' loop this many times
                        jmp     #init                   ' tristate pins and listen for commands


' command T
pasmramtohub            call    #get_values             ' get hubaddr,ramaddr,len and set control pins
                        and     dira,maskP16P31         '%11111111_11111111_00000000_00000000 inputs
                        andn    outa,maskP16            ' memory /rd low
ramtohub_loop           mov     data_16,ina             ' get the data
                        wrword  data_16,hubaddr         ' move data to hub
                        andn    outa,maskP20            ' clock 161 low
                        or      outa,maskP20            ' clock 161 high
                        add     hubaddr,#2              ' increment the hub address 
                        djnz    len,#ramtohub_loop
                        or      outa,maskP16            ' memory /rd high  
                        jmp     #init                   ' ' tristate pins and listen for commands

' command U

pasmramtodisplay        call    #get_values             ' get hubaddr,ramaddr,len (only uses len) and set control pins
                        andn    outa,maskP19            ' CS low
                        and     dira,maskP16P31         ' %11111111_11111111_00000000_00000000 so prop pins 0-15 HiZ
                        andn    outa,maskP16            ' ram /rd low
ramtodisplay_loop       andn    outa,maskP18            ' ILI write low
                        or      outa,maskP18            ' ILI write high
                        andn    outa,maskP20            ' clock 161 low
                        or      outa,maskP20            ' clock 161 high
                        djnz    len,#ramtodisplay_loop
                        or      outa,maskP16            ' mem /rd high
                        or      outa,maskP19            ' CS high 
                        jmp     #init

' command V

pasmhubtodisplay        call    #get_values             ' get hubaddr,ramaddr,len and set control pins high
                        andn    outa,maskP19            ' CS low
                        or      dira,maskP0P15          '%00000000_00000000_11111111_11111111   
hubtodisplay_loop       and     outa,maskP16P31         '%11111111_11111111_00000000_00000000       ' clear for output                   
                        rdword  data_16,hubaddr         ' get the word from hub
                        and     data_16,maskP0P15       ' mask to a word only
                        or      outa,data_16            ' send out the byte to P0-P15
                        andn    outa,maskP18            ' ILI write low
                        or      outa,maskP18            ' ILI write high
                        add     hubaddr,#2              ' one word
                        djnz    len,#hubtodisplay_loop
                        or      outa,maskP19            ' CS high
                        jmp     #init

'command E
RawtoILIformat          ' takes a .raw 3 byte RRRRRRRR GGGGGGGG BBBBBBBB and converts to 2 byte RRRRRGGG GGGBBBBB
                        ' pass hubaddr, ramaddr and len
                        ' hubaddr is source location, len is number of pixels
                        ' ramaddr is destination in hub (messy naming) and length is 2/3 of blocklength
                        call    #get_values ' gets hubaddress, ramaddress and len (ignores ramaddress)
rawloop
                        rdbyte red,hubaddr
                        add hubaddr,#1
                        rdbyte green,hubaddr
                        add hubaddr,#1
                        rdbyte blue,hubaddr
                        add hubaddr,#1
                        call #rgbtoili
                        wrbyte ililow,ramaddr
                        add ramaddr,#1
                        wrbyte ilihigh,ramaddr
                        add ramaddr,#1
                        djnz    len,#rawloop            ' loop until done 
                        jmp     #init                   ' set pins to tristate

RGBtoILI                ' pass red,green, blue, returns ililow and ilihigh
                        shr     red,#3                  ' 000RRRRR 
                        shl     red,#3                  ' RRRRR000 
                        shr     green,#2                ' 00GGGGGG
                        mov     ilihigh,green           ' ilihigh = 00GGGGGG
                        shr     ilihigh,#3              ' ilihigh = 00000GGG
                        or      ilihigh,red             ' ilihigh = RRRRRGGG
                        and     green,#%00000111        ' 00000GGG
                        shl     green,#5                ' GGG00000
                        mov     ililow,green            ' ililow = GGG00000
                        shr     blue,#3                 ' blue = 000BBBBB
                        or      ililow,blue             ' ililow = GGGBBBBB
RGBtoILI_ret            ret

BMPtoILIformat          ' takes a .bmp 3 byte BBBBBBBB GGGGGGGG RRRRRRRR and converts to 2 byte RRRRRGGG GGGBBBBB
                        ' same as E above but BGR instead of RGB
                        ' pass hubaddr, ramaddr and len
                        ' hubaddr is source location, len is number of pixels
                        ' ramaddr is destination in hub (messy naming) and length is 2/3 of blocklength
                        call    #get_values ' gets hubaddress, ramaddress and len (ignores ramaddress)
bmploop
                        rdbyte blue,hubaddr
                        add hubaddr,#1
                        rdbyte green,hubaddr
                        add hubaddr,#1
                        rdbyte red,hubaddr
                        add hubaddr,#1
                        call #rgbtoili
                        wrbyte ililow,ramaddr
                        add ramaddr,#1
                        wrbyte ilihigh,ramaddr
                        add ramaddr,#1
                        djnz    len,#bmploop            ' loop until done
                        jmp     #init                   ' set pins to tristate
' **** command X *********************

MergeIcons              call    #get_values ' gets hubaddress, ramaddress,len which are used here as background,icon,mask
                        mov     pasm_n,#59               ' do a single row
mergeiconsloop          rdbyte  ililow,len                 ' reuse ililow, so this is rdword mask,maskcounter
                        and     ililow,#%11111             ' mask off low 5 bits and use just the blue as this is a grayscale bitmap
                        rdword  red,hubaddr              ' reuse red, so actually this is rdword background,backgroundcounter                        
                        cmp     ililow,#%10000   wc       ' compare if >128 (ie mid level gray)
              if_c      jmp     #mergeskip
                        rdword  green,ramaddr            ' reuse green, so this is rdword iconpixel, iconpixelcounter 
                        wrword  green,hubaddr            ' if replace, then move icon pixel to the background     
mergeskip               add     hubaddr,#2
                        add     ramaddr,#2
                        add     len,#2
                        djnz    pasm_n,#mergeiconsloop            ' loop until done 
                        jmp     #init                   'set pins to tristate 

' *** command Z **********
'set161                  call    #get_values             ' gets ramaddr = the 161 value to set
'                        call    #load161pasm            ' ramaddr to the 161 chips
'                        call    #memorytransfer         ' change to group 1
'                        jmp     #init                        


' *** command N **********

latch373                call    #get_values             ' gets hubaddr,ramaddr,len, and set cotnrol pins - got to here with a spin call
                                                        ' call here from internal pasm routines
localset373latch        or      dira,maskP22            ' and pin 22                
                        or      dira,#%1_11111111       ' enable pins 0-7 and 8 as outputs
                        andn    outa,maskP8             ' P8 low
                        or      outa,maskP22            ' pin 22 high
                        and     outa,maskP0P7low        ' P0-P7 low
                        or      outa,hubaddr            ' send out the data (pass in hubaddr)
                        or      outa,maskP8             ' P8 high, clocks out data
                        andn    outa,maskP8             ' P8 low
                        andn    outa,maskP22            ' pin 22 low
                        jmp     #init


' variables
pasm_n                  long    0                                    ' general purpose value
data_16                 long    0                                    ' general purpose value
ililow                  long    0                                    ' low data byte 
ilihigh                 long    0                                    ' high data byte 
red                     long    0                                    ' red, green blue variables
green                   long    0
blue                    long    0

' constants
Zero                    long    %00000000_00000000_00000000_00000000 ' used in several places
maskP0P2low             long    %11111111_11111111_11111111_11111000 ' P0-P2 low
maskP0P20               long    %00000000_00011111_11111111_11111111 ' P0-P18 enabled for output plus P19,P20    
maskP0P18low            long    %11111111_11111000_00000000_00000000 ' P0-P18 low
maskP16                 long    %00000000_00000001_00000000_00000000 ' pin 16
maskP17                 long    %00000000_00000010_00000000_00000000 ' pin 17
maskP18                 long    %00000000_00000100_00000000_00000000 ' pin 18
maskP19                 long    %00000000_00001000_00000000_00000000 ' pin 19
maskP20                 long    %00000000_00010000_00000000_00000000 ' pin 20
maskP22                 long    %00000000_01000000_00000000_00000000 ' pin 22
maskP16P31              long    %11111111_11111111_00000000_00000000 ' pin 16 to pin 31
maskP0P15               long    %00000000_00000000_11111111_11111111 ' for masking words
maskP16P20              long    %00000000_00011111_00000000_00000000
maskP0P20low            long    %11111111_11100000_00000000_00000000 ' for returning all group pins HiZ
'maskP16P17P19           long    %00000000_00001011_00000000_00000000
maskP16P17P20           long    %00000000_00010011_00000000_00000000
maskP0P7low             long    %11111111_11111111_11111111_00000000  ' P0-P7 low
maskP8                  long    %00000000_00000000_00000001_00000000  ' pin 8
                        fit     496
                  